package dev.flyfish.framework.beans.meta.parser.chain.impl;

import dev.flyfish.framework.beans.meta.parser.BeanPropertyAnnotations;
import dev.flyfish.framework.beans.meta.parser.chain.BeanPropertyAnnotationBatchChain;
import dev.flyfish.framework.beans.meta.parser.chain.BeanPropertyAnnotationChain;
import org.springframework.core.annotation.MergedAnnotation;
import org.springframework.util.CollectionUtils;

import java.lang.annotation.Annotation;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static dev.flyfish.framework.beans.meta.parser.BeanPropertyAnnotations.cast;

/**
 * 简单的批量注解链式实现
 *
 * @param <A> 注解类型
 * @param <T> 实际类型
 */
public class SimpleAnnotationBatchChain<A extends Annotation, T> extends BasicAnnotationChainSupport
        implements BeanPropertyAnnotationBatchChain<A, T> {

    // 内置值
    private List<T> values;

    // 是否map过
    private boolean mapped = false;

    public SimpleAnnotationBatchChain(BeanPropertyAnnotations parent, List<T> annotations) {
        super(parent);
        this.values = annotations;
    }

    /**
     * 映射转换，将转换每一条
     *
     * @param mapper 映射器
     * @return 结果
     */
    @Override
    public <R> BeanPropertyAnnotationBatchChain<A, R> map(Function<T, R> mapper) {
        List<R> list = values.stream().map(mapper).filter(Objects::nonNull)
                .collect(Collectors.toList());
        if (CollectionUtils.isEmpty(list)) {
            return parent.batchEmpty();
        }
        this.mapped = true;
        this.values = cast(list);
        return cast(this);
    }

    /**
     * 过滤，不满足的将过滤，如果为空，将empty
     *
     * @param filter 过滤器
     * @return 结果
     */
    @Override
    public BeanPropertyAnnotationBatchChain<A, T> filter(Predicate<T> filter) {
        this.values = values.stream().filter(filter).collect(Collectors.toList());
        if (CollectionUtils.isEmpty(values)) {
            return parent.batchEmpty();
        }
        return this;
    }

    /**
     * 过滤，不满足的将过滤，如果为空，将empty
     *
     * @param filter 过滤器
     * @return 结果
     */
    @Override
    public BeanPropertyAnnotationBatchChain<A, T> filter(Supplier<Boolean> filter) {
        if (!filter.get()) {
            return parent.batchEmpty();
        }
        return this;
    }

    /**
     * 如果存在，则做
     *
     * @param consumer 消费者
     * @return 结果
     */
    @Override
    public BeanPropertyAnnotationBatchChain<A, T> then(Consumer<List<T>> consumer) {
        this.parent.last(true);
        consumer.accept(this.values);
        return this;
    }

    /**
     * 存在时执行逻辑，不需要结果
     *
     * @param handler 处理器
     * @return 结果
     */
    @Override
    public BeanPropertyAnnotationBatchChain<A, T> exists(Runnable handler) {
        this.parent.last(true);
        handler.run();
        return this;
    }

    /**
     * 不存在做的事情
     *
     * @param fallback 补偿逻辑
     * @return 结果
     */
    @Override
    public BeanPropertyAnnotationBatchChain<A, T> empty(Runnable fallback) {
        return this;
    }

    /**
     * 将多值合并为单值
     *
     * @param reducer 合并器
     * @return 结果
     */
    @Override
    public <R> BeanPropertyAnnotationChain<A, R> reduce(Function<List<T>, R> reducer) {
        R reduced = reducer.apply(values);
        return new SimpleAnnotationChain<>(parent, reduced);
    }

    /**
     * 是否存在值
     *
     * @return 结果
     */
    @Override
    public boolean isPresent() {
        return !CollectionUtils.isEmpty(values);
    }

    /**
     * 获取所有结果为list
     *
     * @return 结果
     */
    @Override
    @SuppressWarnings("unchecked")
    public List<A> synthesize() {
        return mapped ? Collections.emptyList() : this.values.stream().map(item -> (MergedAnnotation<A>) item)
                .map(MergedAnnotation::synthesize).collect(Collectors.toList());
    }

    /**
     * 缓存值逻辑
     *
     * @return 父亲
     */
    @Override
    protected BeanPropertyAnnotations cache() {
        return parent.value(values);
    }
}
