package dev.flyfish.framework.utils;

import org.springframework.beans.factory.annotation.AnnotatedBeanDefinition;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.type.classreading.MetadataReader;
import org.springframework.core.type.filter.AnnotationTypeFilter;
import org.springframework.core.type.filter.AssignableTypeFilter;
import org.springframework.core.type.filter.TypeFilter;
import org.springframework.lang.NonNull;
import org.springframework.util.ClassUtils;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.Collection;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Stream;

/**
 * 内部的包扫描器，提供特定注解扫描
 *
 * @author wangyu
 */
public class ClassPathResourceScanner extends ClassPathScanningCandidateComponentProvider {

    private final TypeFilter filter;

    /**
     * 扫描某个特定注解
     *
     * @param annoType 注解类型
     * @return 结果
     */
    public static ClassPathResourceScanner forAnnotation(Class<? extends Annotation> annoType) {
        return new ClassPathResourceScanner(new AnnotationTypeFilter(annoType));
    }

    /**
     * 扫描子类
     *
     * @param superType 父类型
     * @return 结果
     */
    public static ClassPathResourceScanner forSuperType(Class<?> superType) {
        return new ClassPathResourceScanner(new AssignableTypeFilter(superType));
    }

    private ClassPathResourceScanner(TypeFilter filter) {
        super(false);
        this.filter = filter;
        resetFilters(false);
        addIncludeFilter(filter);
    }

    @Override
    protected boolean isCandidateComponent(@NonNull MetadataReader metadataReader) throws IOException {
        return filter.match(metadataReader, getMetadataReaderFactory());
    }

    @Override
    protected boolean isCandidateComponent(@NonNull AnnotatedBeanDefinition beanDefinition) {
        return true;
    }

    /**
     * 解析具体类型
     *
     * @param bf bean定义
     * @param cl 类加载器
     * @return 结果
     */
    private Class<?> resolveType(BeanDefinition bf, ClassLoader cl) {
        if (null != bf.getBeanClassName()) {
            try {
                return ClassUtils.forName(bf.getBeanClassName(), cl);
            } catch (ClassNotFoundException e) {
                return null;
            }
        }
        return null;
    }

    /**
     * 扫描类
     *
     * @param packageNames 包名
     * @return 结果
     */
    public Stream<Class<?>> scan(Collection<String> packageNames) {
        // 获取扫描器的ClassLoader，保证同源
        ClassLoader cl = this.getClass().getClassLoader();
        Stream<Class<?>> stream = packageNames.stream()
                .flatMap(packageName -> {
                    Set<BeanDefinition> bfs = findCandidateComponents(packageName);
                    // 不存在，不要浪费性能
                    if (CollectionUtils.isEmpty(bfs)) return Stream.empty();
                    // 代理并生成子类，并注册到ioc容器
                    return bfs.stream().map(bf -> resolveType(bf, cl)).filter(Objects::nonNull);
                });
        // 额外返回，以保证类型正确
        return stream.distinct();
    }

    /**
     * 扫描类
     *
     * @param packageNames 包名
     * @return 结果
     */
    public Stream<Class<?>> scan(String... packageNames) {
        return scan(Arrays.asList(packageNames));
    }
}
