package dev.flyfish.framework.utils;

import dev.flyfish.framework.domain.authorized.AuthorizedDomain;
import dev.flyfish.framework.domain.authorized.AuthorizedVo;
import dev.flyfish.framework.domain.base.IUser;
import dev.flyfish.framework.domain.po.Department;
import dev.flyfish.framework.domain.po.Role;
import dev.flyfish.framework.enums.UserType;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.springframework.data.util.CastUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * 部门工具类
 *
 * @author wangyu
 */
public final class DepartUtils {

    /**
     * 合并部门，有包含关系的部门将被合并
     *
     * @param departments 部门
     * @return 结果
     */
    public static Set<String> mergeDeparts(List<Department> departments) {
        if (CollectionUtils.isEmpty(departments)) {
            return Collections.emptySet();
        }
        Set<String> departs = departments.stream().map(Department::getId)
                .collect(Collectors.toSet());
        List<String> removing = new ArrayList<>();
        for (Department department : departments) {
            if (CollectionUtils.isNotEmpty(department.getParentIds()) &&
                    CollectionUtils.containsAny(department.getParentIds(), departs)) {
                removing.add(department.getId());
            }
        }
        removing.forEach(departs::remove);
        return departs;
    }

    /**
     * 包装权限，必须调用
     *
     * @param item 内容
     * @return 结果
     */
    public static <T> T wrapAuthority(T item, AuthorizedDomain po) {
        if (item instanceof AuthorizedVo) {
            AuthorizedVo<?> vo = CastUtils.cast(item);
            // 已发布的内容，谁都不能修改
            vo.setReadonly(false);
            // 获取当前用户
            IUser user = po.getCurrentUser();
            // 非超级管理员，才需要判定
            if (user.getType() != UserType.SUPER_ADMIN) {
                // 用户所属部门
                Set<String> userDeparts = DepartUtils.mergeDeparts(user.getDepartments());
                // 实体归属部门
                String depart = po.getAuthorizeId();
                // 用户权限合集
                Set<Role.Authority> authorities = user.getRoles().stream()
                        .flatMap(role -> null == role.getAuthorities() ? Stream.empty() : role.getAuthorities().stream())
                        .collect(Collectors.toSet());
                // 取出权限，便于判定
                boolean admin = authorities.contains(Role.Authority.ADMIN);
                boolean edit = authorities.contains(Role.Authority.EDIT);
                boolean editChildren = authorities.contains(Role.Authority.EDIT_CHILDREN);
                // 开始判定只读情况，已发布，仅有编辑权限的用户可编辑，管理员权限或者相应部门可编辑权限，才具有读写权限
                if (BooleanUtils.isTrue(po.getPublished())) {
                    vo.setReadonly(!admin && (
                            !edit && userDeparts.contains(depart) || !editChildren && !userDeparts.contains(depart)
                    ));
                } else {
                    // 草稿状态，仅创建者可编辑，其余任何人不可编辑（除了超级管理员）
                    vo.setReadonly(!po.getCreatorId().equals(user.getId()));
                }
            }
        }
        return item;
    }
}
