package org.springframework.data.r2dbc.query;

import dev.flyfish.framework.r2dbc.query.adaptor.FunctionCriteria;
import org.springframework.data.mapping.context.MappingContext;
import org.springframework.data.r2dbc.convert.R2dbcConverter;
import org.springframework.data.r2dbc.dialect.R2dbcDialect;
import org.springframework.data.relational.core.dialect.Escaper;
import org.springframework.data.relational.core.mapping.RelationalPersistentEntity;
import org.springframework.data.relational.core.mapping.RelationalPersistentProperty;
import org.springframework.data.relational.core.query.CriteriaDefinition;
import org.springframework.data.relational.core.query.ValueFunction;
import org.springframework.data.relational.core.sql.*;
import org.springframework.data.relational.core.sql.render.RenderContext;
import org.springframework.data.util.Pair;
import org.springframework.data.util.TypeInformation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.r2dbc.core.Parameter;
import org.springframework.r2dbc.core.binding.BindMarker;
import org.springframework.r2dbc.core.binding.BindMarkers;
import org.springframework.r2dbc.core.binding.MutableBindings;
import org.springframework.util.Assert;

import java.util.*;

/**
 * 全量支持的mapper
 *
 * @author wangyu
 * 低级重写，后续官方支持后删除
 */
@SuppressWarnings({"rawtypes", "unchecked", "deprecation"})
public class FullUpdateMapper extends UpdateMapper {

    private final R2dbcDialect dialect;
    private final RenderContext renderContext;
    private final MappingContext<? extends RelationalPersistentEntity<?>, RelationalPersistentProperty> mappingContext;

    public FullUpdateMapper(R2dbcDialect dialect, R2dbcConverter converter, RenderContext renderContext) {
        super(dialect, converter);
        this.dialect = dialect;
        this.mappingContext = (MappingContext) converter.getMappingContext();
        this.renderContext = renderContext;
    }

    /**
     * 重写此方法，以允许使用自定义函数
     *
     * @param markers  绑定标记对象，必须非空
     * @param criteria 需映射的条件定义，必须非空
     * @param table    必须非空
     * @param entity   相关的 {@link RelationalPersistentEntity}，可以为空
     * @return 结果
     */
    @Override
    @NonNull
    public BoundCondition getMappedObject(@NonNull BindMarkers markers, @NonNull CriteriaDefinition criteria, @NonNull Table table,
                                          @Nullable RelationalPersistentEntity<?> entity) {
        Assert.notNull(markers, "BindMarkers must not be null");
        Assert.notNull(criteria, "CriteriaDefinition must not be null");
        Assert.notNull(table, "Table must not be null");

        MutableBindings bindings = new MutableBindings(markers);
        if (criteria.isEmpty()) {
            throw new IllegalArgumentException("Cannot map empty Criteria");
        }
        Condition mapped = unroll(criteria, table, entity, bindings);
        return new BoundCondition(bindings, mapped);
    }

    private Condition unroll(CriteriaDefinition criteria, Table table, @Nullable RelationalPersistentEntity<?> entity,
                             MutableBindings bindings) {

        CriteriaDefinition current = criteria;

        // reverse unroll criteria chain
        Map<CriteriaDefinition, CriteriaDefinition> forwardChain = new HashMap<>();

        while (current.hasPrevious()) {
            forwardChain.put(current.getPrevious(), current);
            current = current.getPrevious();
        }

        // perform the actual mapping
        Condition mapped = getCondition(current, bindings, table, entity);
        while (forwardChain.containsKey(current)) {

            CriteriaDefinition criterion = forwardChain.get(current);
            Condition result = null;

            Condition condition = getCondition(criterion, bindings, table, entity);
            if (condition != null) {
                result = combine(criterion, mapped, criterion.getCombinator(), condition);
            }

            if (result != null) {
                mapped = result;
            }
            current = criterion;
        }

        if (mapped == null) {
            throw new IllegalStateException("Cannot map empty Criteria");
        }

        return mapped;
    }

    @Nullable
    private Condition unrollGroup(List<? extends CriteriaDefinition> criteria, Table table,
                                  CriteriaDefinition.Combinator combinator, @Nullable RelationalPersistentEntity<?> entity,
                                  MutableBindings bindings) {

        Condition mapped = null;
        for (CriteriaDefinition criterion : criteria) {

            if (criterion.isEmpty()) {
                continue;
            }

            Condition condition = unroll(criterion, table, entity, bindings);

            mapped = combine(criterion, mapped, combinator, condition);
        }

        return mapped;
    }


    private Condition combine(CriteriaDefinition criteria, @Nullable Condition currentCondition,
                              CriteriaDefinition.Combinator combinator, Condition nextCondition) {

        if (currentCondition == null) {
            currentCondition = nextCondition;
        } else if (combinator == CriteriaDefinition.Combinator.INITIAL) {
            currentCondition = currentCondition.and(Conditions.nest(nextCondition));
        } else if (combinator == CriteriaDefinition.Combinator.AND) {
            currentCondition = currentCondition.and(nextCondition);
        } else if (combinator == CriteriaDefinition.Combinator.OR) {
            currentCondition = currentCondition.or(nextCondition);
        } else {
            throw new IllegalStateException("Combinator " + combinator + " not supported");
        }

        return currentCondition;
    }

    @Nullable
    private Condition getCondition(CriteriaDefinition criteria, MutableBindings bindings, Table table,
                                   @Nullable RelationalPersistentEntity<?> entity) {

        if (criteria.isEmpty()) {
            return null;
        }

        if (criteria.isGroup()) {

            Condition condition = unrollGroup(criteria.getGroup(), table, criteria.getCombinator(), entity, bindings);

            return condition == null ? null : Conditions.nest(condition);
        }

        return mapCondition(criteria, bindings, table, entity);
    }


    @Nullable
    private Object convertValue(CriteriaDefinition.Comparator comparator, @Nullable Object value, TypeInformation<?> typeHint) {

        if ((CriteriaDefinition.Comparator.IN.equals(comparator) || CriteriaDefinition.Comparator.NOT_IN.equals(comparator))
                && value instanceof Collection<?> collection && !collection.isEmpty()) {

            Collection<Object> mapped = new ArrayList<>(collection.size());

            for (Object o : collection) {
                mapped.add(convertValue(o, typeHint));
            }

            return mapped;
        }

        return convertValue(value, typeHint);
    }

    /**
     * 核心改造方法，增加对方法标识的适配，剔除额外的column属性
     *
     * @param criteria 通用查询
     * @param bindings 绑定参数
     * @param table    表
     * @param entity   实体信息
     * @return 结果
     */
    private Condition mapCondition(CriteriaDefinition criteria, MutableBindings bindings, Table table,
                                   @Nullable RelationalPersistentEntity<?> entity) {
        Field propertyField = createPropertyField(entity, criteria.getColumn(), mappingContext);
        Column column = table.column(propertyField.getMappedColumnName());
        TypeInformation<?> actualType = propertyField.getTypeHint().getRequiredActualType();

        Object mappedValue;
        Class<?> typeHint;

        CriteriaDefinition.Comparator comparator = criteria.getComparator();
        if (criteria.getValue() instanceof Parameter parameter) {

            mappedValue = convertValue(comparator, parameter.getValue(), propertyField.getTypeHint());
            typeHint = getTypeHint(mappedValue, actualType.getType(), parameter);
        } else if (criteria.getValue() instanceof ValueFunction<?> valueFunction) {

            mappedValue = valueFunction.map(v -> convertValue(comparator, v, propertyField.getTypeHint()))
                    .apply(getEscaper(comparator));

            typeHint = actualType.getType();
        } else {

            mappedValue = convertValue(comparator, criteria.getValue(), propertyField.getTypeHint());
            typeHint = actualType.getType();
        }

        // 特殊处理方法
        if (criteria instanceof FunctionCriteria functionCriteria) {
            return createFunction(column, mappedValue, typeHint, bindings, functionCriteria);
        }
        // 原处理方法
        return createCondition(column, mappedValue, typeHint, bindings, comparator, criteria.isIgnoreCase());
    }

    /**
     * 创建函数引用
     *
     * @param column      引用的列
     * @param mappedValue 映射后的值
     * @param valueType   值类型
     * @param bindings    绑定上下文
     * @param criteria    查询信息
     * @return 结果
     */
    private Condition createFunction(Column column, @Nullable Object mappedValue, Class<?> valueType,
                                     MutableBindings bindings, FunctionCriteria criteria) {
        // 处理多值情况
        if (mappedValue instanceof Iterable) {
            // 大小 + 1， 容纳列名
            List<Expression> expressions = new ArrayList<>(
                    mappedValue instanceof Collection ? ((Collection<?>) mappedValue).size() + 1 : 10);
            expressions.add(column);

            for (Object o : (Iterable<?>) mappedValue) {

                BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());
                expressions.add(bind(o, valueType, bindings, bindMarker));
            }
            return FunctionCondition.of(renderContext, criteria.getFunctionName(), expressions);
        } else {
            // 单值绑定
            BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());
            Expression expression = bind(mappedValue, valueType, bindings, bindMarker);

            return FunctionCondition.of(renderContext, criteria.getFunctionName(), column, expression);
        }
    }

    private Condition createCondition(Column column, @Nullable Object mappedValue, Class<?> valueType,
                                      MutableBindings bindings, CriteriaDefinition.Comparator comparator, boolean ignoreCase) {

        if (comparator.equals(CriteriaDefinition.Comparator.IS_NULL)) {
            return column.isNull();
        }

        if (comparator.equals(CriteriaDefinition.Comparator.IS_NOT_NULL)) {
            return column.isNotNull();
        }

        if (comparator == CriteriaDefinition.Comparator.IS_TRUE) {
            Expression bind = booleanBind(column, mappedValue, valueType, bindings, ignoreCase);

            return column.isEqualTo(bind);
        }

        if (comparator == CriteriaDefinition.Comparator.IS_FALSE) {
            Expression bind = booleanBind(column, mappedValue, valueType, bindings, ignoreCase);

            return column.isEqualTo(bind);
        }

        Expression columnExpression = column;
        if (ignoreCase) {
            columnExpression = Functions.upper(column);
        }

        if (comparator == CriteriaDefinition.Comparator.NOT_IN || comparator == CriteriaDefinition.Comparator.IN) {

            Condition condition;

            if (mappedValue instanceof Iterable) {

                List<Expression> expressions = new ArrayList<>(
                        mappedValue instanceof Collection ? ((Collection<?>) mappedValue).size() : 10);

                for (Object o : (Iterable<?>) mappedValue) {

                    BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());
                    expressions.add(bind(o, valueType, bindings, bindMarker));
                }

                condition = Conditions.in(columnExpression, expressions.toArray(new Expression[0]));

            } else {

                BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker);

                condition = Conditions.in(columnExpression, expression);
            }

            if (comparator == CriteriaDefinition.Comparator.NOT_IN) {
                condition = condition.not();
            }

            return condition;
        }

        if (comparator == CriteriaDefinition.Comparator.BETWEEN || comparator == CriteriaDefinition.Comparator.NOT_BETWEEN) {

            Pair<Object, Object> pair = (Pair<Object, Object>) mappedValue;

            Expression begin = bind(pair.getFirst(), valueType, bindings,
                    bindings.nextMarker(column.getName().getReference()), ignoreCase);
            Expression end = bind(pair.getSecond(), valueType, bindings, bindings.nextMarker(column.getName().getReference()),
                    ignoreCase);

            return comparator == CriteriaDefinition.Comparator.BETWEEN ? Conditions.between(columnExpression, begin, end)
                    : Conditions.notBetween(columnExpression, begin, end);
        }

        BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());

        switch (comparator) {
            case EQ: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker, ignoreCase);
                return Conditions.isEqual(columnExpression, expression);
            }
            case NEQ: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker, ignoreCase);
                return Conditions.isEqual(columnExpression, expression).not();
            }
            case LT: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker);
                return column.isLess(expression);
            }
            case LTE: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker);
                return column.isLessOrEqualTo(expression);
            }
            case GT: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker);
                return column.isGreater(expression);
            }
            case GTE: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker);
                return column.isGreaterOrEqualTo(expression);
            }
            case LIKE: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker, ignoreCase);
                return Conditions.like(columnExpression, expression);
            }
            case NOT_LIKE: {
                Expression expression = bind(mappedValue, valueType, bindings, bindMarker, ignoreCase);
                return Conditions.notLike(columnExpression, expression);
            }
            default:
                throw new UnsupportedOperationException("Comparator " + comparator + " not supported");
        }
    }

    private Expression booleanBind(Column column, Object mappedValue, Class<?> valueType, MutableBindings bindings,
                                   boolean ignoreCase) {
        BindMarker bindMarker = bindings.nextMarker(column.getName().getReference());

        return bind(mappedValue, valueType, bindings, bindMarker, ignoreCase);
    }

    private Expression bind(@Nullable Object mappedValue, Class<?> valueType, MutableBindings bindings,
                            BindMarker bindMarker) {
        return bind(mappedValue, valueType, bindings, bindMarker, false);
    }

    private Expression bind(@Nullable Object mappedValue, Class<?> valueType, MutableBindings bindings,
                            BindMarker bindMarker, boolean ignoreCase) {

        if (mappedValue != null) {
            bindings.bind(bindMarker, mappedValue);
        } else {
            bindings.bindNull(bindMarker, valueType);
        }

        return ignoreCase ? Functions.upper(SQL.bindMarker(bindMarker.getPlaceholder()))
                : SQL.bindMarker(bindMarker.getPlaceholder());
    }


    private Escaper getEscaper(CriteriaDefinition.Comparator comparator) {

        if (comparator == CriteriaDefinition.Comparator.LIKE || comparator == CriteriaDefinition.Comparator.NOT_LIKE) {
            return dialect.getLikeEscaper();
        }

        return Escaper.DEFAULT;
    }

}
