package dev.flyfish.framework.file.utils;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.util.StreamUtils;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.zip.ZipFile;

/**
 * zip解压缩工具
 *
 * @author wangyu
 */
@Slf4j
public class ZipUtils {

    /**
     * 解压到当前目录
     *
     * @param path 路径
     */
    public static List<String> unzip(String path) throws IOException {
        // 目录
        String directory = StringUtils.substringBeforeLast(path, "/");
        // 解压
        return unzip(path, directory);
    }

    /**
     * 解压文件到指定目录
     * 解压后的文件名，和之前一致
     *
     * @param path   zip文件目录
     * @param target 目标解压目录
     */
    public static List<String> unzip(String path, String target) throws IOException {
        // 解决中文文件夹乱码
        ZipFile zip = new ZipFile(path, Charset.defaultCharset());
        // 取得真实名字
        String name = StringUtils.substringBefore(StringUtils.substringAfterLast(path, "/"), ".");
        // 父目录
        String parentPath = target + "/" + name;
        Path pathFile = Paths.get(parentPath);
        if (!Files.exists(pathFile)) {
            Files.createDirectories(pathFile);
        }

        // 循环zip内的文件
        List<String> files = zip.stream().map(entry -> {
            String entryName = entry.getName();
            String output;
            try (InputStream in = zip.getInputStream(entry)) {
                OutputStream out = prepare(output = parentPath + "/" + entryName);
                // 拷贝所有流
                StreamUtils.copy(in, out);
                return output;
            } catch (IOException e) {
                log.error("[zip]解压某个文件发生异常：{}", e.getMessage());
            }
            return null;
        }).filter(Objects::nonNull).collect(Collectors.toList());
        log.debug("解压了:{}", files);
        return files;
    }

//    public static void main(String[] args) throws IOException {
//        unzip("/Users/wangyu/Documents/ceshi.m3u8");
//    }

    /**
     * output
     * 准备安全的输出环境
     *
     * @param output 输出路径
     * @return 结果
     */
    private static FileOutputStream prepare(String output) throws IOException {
        // 输出文件路径信息
        log.info("抽取{}", output);
        // 判断路径是否存在,不存在则创建文件路径
        Path targetPath = Paths.get(StringUtils.substringBeforeLast(output, "/"));
        if (!Files.exists(targetPath)) {
            Files.createDirectories(targetPath);
        }
        // 判断文件全路径是否为文件夹,如果是上面已经上传,不需要解压
        if (Files.isDirectory(Paths.get(output))) {
            throw new IOException("[unzip]文件路径为目录，跳过写入");
        }
        return new FileOutputStream(output);
    }
}

