package dev.flyfish.framework.user.service;


import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.domain.base.IUser;
import dev.flyfish.framework.domain.base.Qo;
import dev.flyfish.framework.domain.po.Permission;
import dev.flyfish.framework.domain.po.Role;
import dev.flyfish.framework.enums.RoleType;
import dev.flyfish.framework.enums.UserType;
import dev.flyfish.framework.query.Queries;
import dev.flyfish.framework.query.Query;
import dev.flyfish.framework.repository.ReactiveEntityOperations;
import dev.flyfish.framework.service.impl.BaseReactiveServiceImpl;
import dev.flyfish.framework.user.domain.PermissionQo;
import dev.flyfish.framework.user.domain.RoleQo;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.Collections;
import java.util.Set;
import java.util.stream.Collectors;

@Service
@RequiredArgsConstructor
public class RoleService extends BaseReactiveServiceImpl<Role> {

    private final PermissionService permissionService;

    private final ReactiveEntityOperations reactiveEntityOperations;

    /**
     * 如果是管理员，设置拥有所有权限
     *
     * @param role 实体
     * @return 结果
     */
    @Override
    public Mono<Role> create(Role role) {
        return Mono.defer(() -> {
                    if (BooleanUtils.isTrue(role.getSystem())) {
                        return permissionService.getAll().collectList();
                    } else if (BooleanUtils.isTrue(role.getAdmin())) {
                        // 如果是管理员，拥有所有权限
                        PermissionQo qo = new PermissionQo();
                        qo.setAdmin(false);
                        return permissionService.getList(qo).collectList();
                    }
                    return Mono.just(Collections.<Permission>emptyList());
                })
                .flatMap(permissions -> {
                    role.setPermissions(permissions);
                    if (null == role.getType()) {
                        role.setType(RoleType.PC);
                    }
                    return super.create(role);
                });
    }

    /**
     * 获取创建的主键集合
     *
     * @param user 用户
     * @return 结果
     */
    private Mono<Set<String>> getOwnedIds(IUser user) {
        Query query = Query.query(Queries.where(Role::getCreatorId).eq(user.getId())).select(Role::getId);
        return reactiveEntityOperations.findAll(query, this.getEntityInformation()).map(Domain::getId).collect(Collectors.toSet());
    }

    /**
     * 查询时，根据用户过滤可见项
     *
     * @param query 查询
     * @return 结果
     */
    @Override
    public Flux<Role> getList(Qo<Role> query) {
        // 基于创建者进行筛选
        if (query instanceof RoleQo) {
            IUser user = query.getUser();
            if (null != user && user.getType() != UserType.SUPER_ADMIN) {
                return getOwnedIds(user)
                        .flatMapMany(ids -> {
                            RoleQo qo = (RoleQo) query;
                            qo.setIds(ids);
                            if (CollectionUtils.isNotEmpty(qo.getAdditions())) {
                                qo.getIds().addAll(qo.getAdditions());
                            }
                            if (CollectionUtils.isEmpty(qo.getIds())) {
                                return Flux.empty();
                            }
                            return super.getList(qo);
                        });
            }
        }
        return super.getList(query);
    }
}
