package dev.flyfish.framework.controller.reactive;

import dev.flyfish.framework.annotations.Operation;
import dev.flyfish.framework.bean.Result;
import dev.flyfish.framework.bean.SyncVo;
import dev.flyfish.framework.configuration.annotations.PagedQuery;
import dev.flyfish.framework.configuration.annotations.ValidRequestBody;
import dev.flyfish.framework.controller.SafeController;
import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.domain.base.Qo;
import dev.flyfish.framework.service.BaseReactiveService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.util.CastUtils;
import org.springframework.web.bind.annotation.*;
import reactor.core.publisher.Mono;

import java.util.List;

/**
 * 通用RESTful的控制器，用于动态提供基础的RESTful接口
 * 基于最新知识架构出品
 *
 * @author wangyu
 * @create 2021-12-7
 */
public abstract class ReactiveBaseController<T extends Domain, Q extends Qo<T>> implements SafeController {

    @Autowired
    protected BaseReactiveService<T> reactiveService;

    public <S extends BaseReactiveService<T>> S getService() {
        return CastUtils.cast(reactiveService);
    }

    @GetMapping("/exists")
    public Mono<Result<Boolean>> exists(@PagedQuery Q qo) {
        return reactiveService.exists(qo).map(Result::accept);
    }

    @GetMapping("/count")
    public Mono<Result<Long>> count(@PagedQuery Q qo) {
        return reactiveService.count(qo).map(Result::accept);
    }

    @PostMapping(value = "", headers = "Draft=1")
    @Operation.Create
    public Mono<Result<T>> createDraft(@RequestBody T entity) {
        return reactiveService.create(entity).map(Result::accept);
    }

    @PostMapping("")
    @Operation.Create
    public Mono<Result<T>> create(@ValidRequestBody T entity) {
        return reactiveService.create(entity).map(Result::accept);
    }

    @GetMapping("{id}")
    public Mono<Result<T>> get(@PathVariable String id) {
        return reactiveService.getById(id).map(Result::accept).defaultIfEmpty(Result.notFound());
    }

    @PutMapping(value = "{id}", headers = "Draft=1")
    @Operation.Update
    public Mono<Result<T>> updateDraft(@RequestBody T entity) {
        return reactiveService.updateSelectiveById(entity).map(Result::accept).defaultIfEmpty(Result.notFound());
    }

    @PutMapping("{id}")
    @Operation.Update
    public Mono<Result<T>> update(@ValidRequestBody T entity) {
        return reactiveService.updateSelectiveById(entity).map(Result::accept).defaultIfEmpty(Result.notFound());
    }

    @PatchMapping("{id}")
    @Operation.Update
    public Mono<Result<T>> patch(@RequestBody T entity) {
        return reactiveService.updateSelectiveById(entity).map(Result::accept).defaultIfEmpty(Result.notFound());
    }

    @PutMapping("")
    @Operation.UpdateAll
    public Mono<Result<List<T>>> updateList(@RequestBody List<T> list) {
        return reactiveService.updateBatch(list).collectList().map(Result::accept).defaultIfEmpty(Result.emptyList());
    }

    @PutMapping("/sync")
    @Operation.Sync
    public Mono<Result<SyncVo<T>>> syncList(@RequestBody List<T> list) {
        return reactiveService.sync(list).map(Result::accept).defaultIfEmpty(Result.accept(SyncVo.<T>builder().build()));
    }

    @DeleteMapping("{id}")
    @Operation.Delete
    public Mono<Result<T>> remove(@PathVariable List<String> id) {
        return reactiveService.deleteBatchByIds(id).thenReturn(Result.ok());
    }

    @GetMapping("/all")
    public Mono<Result<List<T>>> all(@PagedQuery Q qo) {
        if (qo.isEmpty()) {
            return reactiveService.getAll().collectList().map(Result::accept).defaultIfEmpty(Result.emptyList());
        }
        return reactiveService.getList(qo).collectList().map(Result::accept).defaultIfEmpty(Result.emptyList());
    }

    @GetMapping("")
    public Mono<Result<List<T>>> list(@PagedQuery Q qo) {
        if (null != qo.getPageable() && qo.getPageable().isPaged()) {
            return reactiveService.getPageList(qo).map(Result::accept).defaultIfEmpty(Result.emptyPage());
        }
        return reactiveService.getList(qo).collectList().map(Result::accept).defaultIfEmpty(Result.emptyList());
    }
}
